/***************************************************************************
 *   Copyright (C) 2009 by Stanson <stanson@pangolin.ru>                   *
 *   GPL v3 code, read COPYING file                                        *
 **************************************************************************/
#define extern
#include <pic18fregs.h>
#undef extern
#include "usb.h"
#include "config.h"
#include "main.h"

data at (0x408) BDT ep1Bo;		//Endpoint #1 BD Out
data at (0x40C) BDT ep1Bi;		//Endpoint #1 BD In

#pragma udata bank5 dataIN
byte dataIN[OSC_EP_SIZE];
#pragma udata bank6 dataOUT
byte dataOUT[OSC_EP_SIZE];
#pragma udata bank7 osc_data
byte osc_data[256];

byte cnt, cnt2;
byte delay;
byte timescale; /* 4.2 us, 10.6 us, 20 us, 50 us, 0.1 ms, 0.2 ms, 0.5 ms */
byte sync;
byte v1, v2;

extern byte           usb_device_state;	// Device States: DETACHED, ATTACHED, ...

#pragma stack 0x200 255

#pragma code _reset 0x800
void _reset (void) _naked
{
    /*       */
    _asm
    LFSR 1, _stack_end
    GOTO _main
    _endasm;
}

#pragma code _high_ISR 0x808
void _high_ISR (void) _naked
{
    _asm
    GOTO _high_interrupt
    _endasm;
}

#pragma code _low_ISR 0x818
void _low_ISR (void) _naked
{
    _asm
    GOTO _low_interrupt
    _endasm;
}


void high_interrupt(void) interrupt
{
}

void low_interrupt(void) interrupt
{
}

#define GET_8BYTES \
      _asm 			MOVFF	_PORTB, _POSTINC0 	_endasm; \
      _asm 			MOVFF	_PORTB, _POSTINC0 	_endasm; \
      _asm 			MOVFF	_PORTB, _POSTINC0 	_endasm; \
      _asm 			MOVFF	_PORTB, _POSTINC0 	_endasm; \
      _asm 			MOVFF	_PORTB, _POSTINC0 	_endasm; \
      _asm 			MOVFF	_PORTB, _POSTINC0 	_endasm; \
      _asm 			MOVFF	_PORTB, _POSTINC0 	_endasm; \
      _asm 			MOVFF	_PORTB, _POSTINC0 	_endasm;


#define GET_32BYTES \
    GET_8BYTES \
    GET_8BYTES \
    GET_8BYTES \
    GET_8BYTES

void osc_get_data(void)
{
  byte delta = 0, dv = 0;

  /*
   * cycles = 2 + 2 + 1 + (1 + 2) * (delay - 1) + 1 + 1 + 2 = 6 + 3 * delay
   * delay = (cycles - 6) / 3
   * cycles = (scale / 25.6) * 12 MHz
   * delay = (scale * 468750 - 6) / 3 = 156250 * scale - 2
   * scale = (delay + 2) / 156250
   */
  switch(timescale)
  {
    case 0: delay = 1;   break;
    case 1: delay = 1;   break;
    case 2: delay = 1;   break; /*  20 uS/div (19.2  uS/div) */
    case 3: delay = 6;   break; /*  50 uS/div (51.2  uS/div) */
    case 4: delay = 13;  break; /* 0.1 mS/div (0.096 mS/div) */
    case 5: delay = 29;  break; /* 0.2 mS/div (0.198 mS/div) */
    case 6: delay = 76;  break; /* 0.5 mS/div (0.499 mS/div) */
    case 7: delay = 154; break; /*   1 mS/div (0.998 mS/div) */
    case 8: delay = 0;   break; /*   2 mS/div (1.65  ms/div) */
    default: delay = 0;  break;
  }

  if(sync != 128)
  {
    cnt = 0;
    if(sync > 128)
    {
      delta = sync - 128;
      while( dv < delta && --cnt )
      {
        _asm 			MOVFF	_PORTB, _v1	 	_endasm;
        _asm			MOVFF	_delay, _cnt2		_endasm; /* 2 c */
        _asm _osc_get_data01:	DECFSZ	_cnt2, F, B 		_endasm; /* 1 c */
        _asm 			BRA	_osc_get_data01		_endasm; /* 2 c */
        _asm 			MOVFF	_PORTB, _v2	 	_endasm;
        dv = (v2 > v1) ?  v2 - v1 : 0;
      }
    }
    else
    {
      delta = 128 - sync;
      while( dv < delta && --cnt )
      {
        _asm 			MOVFF	_PORTB, _v2	 	_endasm;
        _asm			MOVFF	_delay, _cnt2		_endasm; /* 2 c */
        _asm _osc_get_data02:	DECFSZ	_cnt2, F, B 		_endasm; /* 1 c */
        _asm 			BRA	_osc_get_data02		_endasm; /* 2 c */
        _asm 			MOVFF	_PORTB, _v1	 	_endasm;
        dv = (v2 > v1) ?  v2 - v1 : 0;
      }
    }
  }

  switch(timescale)
  {
    case 0: /* 2 cycles, 6 Ms/Sec, 4.2 uS/div */
      _asm			LFSR	0, _osc_data 		_endasm;
      GET_32BYTES
      GET_32BYTES
      GET_32BYTES
      GET_32BYTES
      GET_32BYTES
      GET_32BYTES
      GET_32BYTES
      GET_32BYTES
      break;

    case 1:
      /* 5 cycles, 2.4 Ms/sec, 10.6 uS/div */
      cnt = 0;
      _asm 			LFSR	0, _osc_data 		_endasm;
      _asm _osc_get_data1:	MOVFF	_PORTB, _POSTINC0 	_endasm; /* 2 c */
      _asm 			DECFSZ	_cnt, F, B 		_endasm; /* 1 c */
      _asm 			BRA	_osc_get_data1 		_endasm; /* 2 c */
      break;

    default:
      /* 9 cycles, 1.3 Ms/sec, 19.2 uS/div */
      cnt = 0;
      _asm 			LFSR	0, _osc_data 		_endasm;
      _asm _osc_get_data3:	MOVFF	_PORTB, _POSTINC0 	_endasm; /* 2 c */

      _asm			MOVFF	_delay, _cnt2		_endasm; /* 2 c */
      _asm _osc_get_data4:	DECFSZ	_cnt2, F, B 		_endasm; /* 1 c */
      _asm 			BRA	_osc_get_data4 		_endasm; /* 2 c */

      _asm 			DECFSZ	_cnt, F, B 		_endasm; /* 1 c */
      _asm 			BRA	_osc_get_data3 		_endasm; /* 2 c */
      break;
  }
}

void OscService(void)
{ byte i, part;

    if((usb_device_state < CONFIGURED_STATE) || (UCONbits.SUSPND)) return;

    if(!(ep1Bo.Stat & _USIE))
    { //  ,  
        switch(dataOUT[0])
        {
            case READ_VERSION:
	        i = 2;
                dataIN[0] = MINOR_VERSION;
                dataIN[1] = MAJOR_VERSION;
                break;

	    case COLLECT_OSC_DATA:
		i = 1;
		osc_get_data();
		dataIN[0] = 1;
		break;

            case READ_OSC_DATA:
		part = dataOUT[1] * 64;
		for(i = 0; i < 64; i++, part++)
		{
		  dataIN[i] = osc_data[part];
		}
                i = 64;
	    	break;

	    case SET_OFFSET:
		CVRCON = 0xC0 | (dataOUT[1] & 0x0F);
		i = 0;
		break;

	    case SET_TIMESCALE:
		timescale = dataOUT[1];
		i = 0;
		break;

	    case SET_SYNC:
		sync = dataOUT[1];
		i = 0;
		break;

	    default:
	        i = 0;
	        break;
        }
	
        if(i != 0)
        { //   
            ep1Bi.Cnt = i;

	    if(ep1Bi.Stat & _DTSMASK)
	        ep1Bi.Stat = _USIE | _DAT0 | _DTSEN;
            else
	        ep1Bi.Stat = _USIE | _DAT1 | _DTSEN;
        }

        //   
        ep1Bo.Cnt = sizeof(dataOUT);

        if(ep1Bo.Stat & _DTSMASK)
            ep1Bo.Stat = _USIE | _DAT0 | _DTSEN;
        else
            ep1Bo.Stat = _USIE | _DAT1 | _DTSEN;

    }
}

void USBUserInit()
{
    UEP1 = EP_OUT_IN | HSHK_EN;           // Enable 2 data pipes

    ep1Bo.Cnt  = OSC_EP_SIZE;              // buffer size
    ep1Bo.ADR  = (data byte*)&dataOUT;   // buffer address
    ep1Bo.Stat = _USIE | _DAT0 | _DTSEN;  // status

    ep1Bi.ADR  = (data byte*)&dataIN;     // buffer address
    ep1Bi.Stat = _UCPU | _DAT1;           // status
}

void main(void)
{
    /*   */

    /*
     * RA0 <- ???
     * RA1 <- ???
     * RA2 <- A-> CVref
     * RA3 <- ???
     * RA4 -> ???
     * RA5 -> USB_OK LED
     */
    PORTA = 0;
    TRISA = 0x0F;

    CMCON = 0x07;
    CVRCON = 0xE7;

    ADCON1 = 0x0F;

    /*
     * RB0-RB7 <- ADC data
     */
    PORTB = 0;
    TRISB = 0xFF;

    /*
     * RC0 <- ???
     * RC1 <- ???
     * RC2 -> CCP1 PWM (ADC Clk)
     * RC6 -> ???
     * RC7 -> ???
     */
    PORTC = 0;
    TRISC = 0x3B;

    /*   (16 , CLK 3  (48  / 4 / 4)) */
//    T0CON = 0x81;
//    INTCONbits.TMR0IF = 0;

    /* ADC clock. PWM1 module. 12MHz output. High on Q1 cycle only */
    CCP1CON = 0x1F;
    CCPR1L  = 0x00;
    PR2     = 0x00;
    TMR2    = 0;
    /*  Timer2 - post: 1:1, pre: 1:1 */
    T2CON = 0x04;

    timescale = 1;    

//    p_count   = 0;
//    p_delay.w = -250; /* 3 000 000 Hz / 12 000 pps = 250 */

//    TMR0H = p_delay.h;
//    TMR0L = p_delay.l;
    

    /*  USB */
    USBInit();

//    PIE2bits.USBIE = 1;            /*   USB */
//    INTCON = 0xC0;                 /*   */

    while(1)
    {
        USBIntService();
        OscService();
    }
}
